import adaptive
import kwant
import kwant.continuum
import numpy as np
from functools import lru_cache
from scipy import integrate


s0 = np.eye(2)
sz = np.array([[1, 0], [0, -1]])
sy = np.array([[0, -1j], [1j, 0]])
sx = np.array([[0, 1], [1, 0]])


TEMPLATE = """(0.5 * hbar**2 * k_x**2 / m_eff * c - mu + V) * kron(sigma_z, sigma_0) +
               alpha * k_x * kron(sigma_z, sigma_y) +
               B * kron(sigma_0, sigma_x) +
               Delta * kron(sigma_x, sigma_0)"""


def barrier_potential(x, d_barrier, asym, V_bias, V_barrier):
    return V_barrier - asym + (asym - V_bias) * (1 - x / d_barrier)


def one_lead_bias_potential(x, d_barrier, asym, V_bias, V_barrier):
    # Normal lead
    if x <= 0:
        return -V_bias

    # Barrier
    elif 0 < x <= d_barrier:
        return barrier_potential(x, d_barrier, asym, V_bias, V_barrier)

    # Proximitized wire
    else:
        return 0


def double_barrier_potential(
    x, d_barrier, L_sc, L_dot, asym, V_bias_left, V_bias_right, V_barrier, right_dot
):
    wire_end = d_barrier + L_dot + L_sc
    if right_dot:
        wire_end += L_dot

    # Left barrier
    if 0 < x <= d_barrier:
        return V_barrier - asym + (asym - V_bias_left) * (1 - x / d_barrier)

    # Right barrier
    elif wire_end < x <= wire_end + d_barrier:
        return V_barrier - asym + (asym - V_bias_right) * (x - wire_end) / d_barrier


def two_lead_bias_potential(
    x, d_barrier, L_sc, L_dot, asym, V_bias_left, V_bias_right, V_barrier, right_dot
):
    wire_end = d_barrier + L_dot + L_sc
    if right_dot:
        wire_end += L_dot

    # Left normal lead
    if x <= 0:
        return -V_bias_left

    # Left barrier
    elif 0 < x <= d_barrier:
        return double_barrier_potential(
            x,
            d_barrier,
            L_sc,
            L_dot,
            asym,
            V_bias_left,
            V_bias_right,
            V_barrier,
            right_dot,
        )

    # Proximitized wire
    elif d_barrier < x <= wire_end:
        return 0

    # Right Barrier
    elif wire_end < x <= wire_end + d_barrier:
        return double_barrier_potential(
            x,
            d_barrier,
            L_sc,
            L_dot,
            asym,
            V_bias_left,
            V_bias_right,
            V_barrier,
            right_dot,
        )

    # Right normal lead
    if x > wire_end + d_barrier:
        return -V_bias_right


def dot_potential(x, d_barrier, V_dot, L_sc, L_dot, right_dot, dot_asym):
    wire_end = d_barrier + L_dot + L_sc
    if right_dot:
        wire_end += L_dot

    if d_barrier < x <= d_barrier + L_dot:
        return V_dot * np.cos(np.pi / 2 + np.pi * (x - d_barrier) / L_dot)
    elif wire_end - L_dot < x <= wire_end:
        _x = x - wire_end
        return (1 - dot_asym) * V_dot * np.cos(np.pi / 2 + np.pi * (_x - L_dot) / L_dot)
    return 0


@lru_cache()
def finite_bias_NS(
    L_norm=np.inf,
    L_sc=np.inf,
    d_barrier=3,
    L_dot=0,
    finalized=True,
    right_lead=False,
    right_dot=False,
    grid=1,
):
    """Make a normal-superconductor barrier, with a potential barrier in the middle
    of width d_barrier. A bias can be applied to the normal part, which then decays
    linearly over the barrier.

    Parameters
    -----------
    L_norm: int
        Length of the normal terminal
    L_sc: int
        Length of the proximitized wire.
    d_barrier: int
        Width of the tunnel barrier in units of the lattice constant.
    L_dot: int
        Width of the quantum dot in units of the lattice constant.
    finalized: bool
        Whether or not to return a finalized system.
    right_lead: bool
        Whether to include a second normal lead to the right
        of the superconductor. Only usable if L_sc is finite.
    """

    Hamiltonian = TEMPLATE

    def interval(site, lower, upper):
        (x,) = site.pos
        return lower <= x < upper

    wire_end = d_barrier + L_dot + L_sc
    if right_dot:
        wire_end += L_dot

    # Scattering region
    # Potential barrier at the interface between the normal lead and the superconductor.
    barrier = kwant.Builder()
    if right_lead:
        barrier_subs = {
            "V": "bias_potential(x, d_barrier, L_sc, L_dot, asym, V_bias_left, V_bias_right, V_barrier, right_dot)",
            "Delta": 0,
        }
    else:
        barrier_subs = {
            "V": "bias_potential(x, d_barrier, asym, V_bias_left, V_barrier)",
            "Delta": 0,
        }

    if d_barrier > 0:
        barrier_template = kwant.continuum.discretize(
            Hamiltonian, grid=grid, locals=barrier_subs
        )
        barrier.fill(
            barrier_template,
            lambda site: interval(site, 0, d_barrier),
            (0,),
            max_sites=float("inf"),
        )

    # Quantum dot potential
    if L_dot > 0:
        dot_subs = {
            "V": "dot_potential(x, d_barrier, V_dot, L_sc, L_dot, right_dot, dot_asym)",
            "Delta": 0,
        }
        dot_template = kwant.continuum.discretize(
            Hamiltonian, grid=grid, locals=dot_subs
        )
        barrier.fill(
            dot_template,
            lambda site: interval(site, d_barrier, d_barrier + L_dot),
            (d_barrier,),
        )

    # Normal lead
    N_subs = {"V": "-V_bias_left", "Delta": 0, "mu": "mu_lead"}
    N_template = kwant.continuum.discretize(Hamiltonian, grid=grid, locals=N_subs)
    if L_norm == np.inf:
        # Separate particle and hole degrees of freedom, make sure the basis is particle-hole symmetric.
        N_lead = kwant.Builder(
            kwant.TranslationalSymmetry((-1,)),
            conservation_law=-np.kron(sz, s0),
            particle_hole=np.kron(sy, sy),
        )
        N_lead.fill(N_template, lambda site: True, (0,), max_sites=float("inf"))
        barrier.attach_lead(N_lead)
    elif L_norm > 0:
        barrier.fill(N_template, lambda site: interval(site, -L_norm, 0), (-L_norm,))

    # Proximitized wire
    SC_subs = {"V": 0}
    SC_template = kwant.continuum.discretize(Hamiltonian, grid=grid, locals=SC_subs)
    if L_sc == np.inf:
        SC_lead = kwant.Builder(kwant.TranslationalSymmetry((1,)))
        SC_lead.fill(SC_template, lambda site: True, (0,), max_sites=float("inf"))
        barrier.attach_lead(SC_lead)
    else:
        barrier.fill(
            SC_template,
            lambda site: interval(site, d_barrier + L_dot, d_barrier + L_dot + L_sc),
            (d_barrier + L_dot,),
        )

    # Right quantum dot
    if L_dot > 0 and right_dot:
        barrier.fill(
            dot_template,
            lambda site: interval(site, wire_end - L_dot, wire_end),
            (wire_end - L_dot,),
        )

    # Right normal lead
    if right_lead:
        if L_sc == np.inf:
            raise ValueError("Right normal lead only compatible with a finite L_sc.")

        barrier_template = kwant.continuum.discretize(
            Hamiltonian, grid=grid, locals=barrier_subs
        )
        barrier.fill(
            barrier_template,
            lambda site: interval(site, wire_end, wire_end + d_barrier),
            (wire_end,),
            max_sites=float("inf"),
        )

        N_subs = {"V": "-V_bias_right", "Delta": 0, "mu": "mu_lead"}
        N_template = kwant.continuum.discretize(Hamiltonian, grid=grid, locals=N_subs)
        if L_norm == np.inf:
            N_lead = kwant.Builder(
                kwant.TranslationalSymmetry((1,)),
                conservation_law=-np.kron(sz, s0),
                particle_hole=np.kron(sy, sy),
            )
            N_lead.fill(N_template, lambda site: True, (0,), max_sites=float("inf"))
            barrier.attach_lead(N_lead)
        elif L_norm > 0:
            barrier.fill(
                N_template,
                lambda site: interval(
                    site, wire_end + d_barrier, wire_end + d_barrier + L_norm
                ),
                (wire_end + d_barrier,),
            )

    if finalized:
        return barrier.finalized()
    else:
        return barrier


def fermi_dist(E, T):
    """Fermi distribution at energy E and temperature T.
    The temperature should be provided in the same units as E, i.e.
    in units of energy.

        E: energy
        T: temperature
       """
    # Treat small values of T as zero.
    if T > 1e-8:
        return 1 / (1 + np.exp(E / T))
    else:
        return 1 if E < 0 else 0


def andreev_conductance(energy, syst, params, in_lead=0, out_lead=1, nonloc=False):
    """Conductance of an NS or NSN junction at energy E in the linear
    response limit at zero temperature.

    Parameters:
    ----------
    energy: float
        Energy at which to compute the conductance.
    syst: kwant.builder.FiniteSystem
        Finalized Kwant system for the superconducting
        heterostructure.
    params: dict
        System parameters to pass to Kwant solvers.
    in_lead: int
        Normal lead from which electrons are injected, either
        0 or 1.
    out_lead: int
        Normal lead where the conductance is computed, either
        0 or 1.
    nonloc: bool
        Whether to compute the local (False) or nonlocal (True)
        conductance.
    """
    # Include the energy in the parameters to make
    # the current computation work in the self-energy
    # implementation
    params["E"] = energy
    smatrix = kwant.smatrix(syst, energy, params=params)
    # Local conductance
    if not nonloc:
        # Conductance is N - R_ee + R_he
        cond = (
            smatrix.submatrix((in_lead, 0), (in_lead, 0)).shape[0]
            - smatrix.transmission((in_lead, 0), (in_lead, 0))
            + smatrix.transmission((in_lead, 1), (in_lead, 0))
        )
    # Nonlocal conductance
    else:
        # Following convention in arXiv:1905.05438
        cond = -smatrix.transmission(
            (in_lead, 0), (out_lead, 0)
        ) + smatrix.transmission((in_lead, 1), (out_lead, 0))
    return cond


def NS_current(
    V_bias,
    syst,
    params,
    nonloc=False,
    in_lead=0,
    out_lead=1,
    biased_lead=0,
    epsabs=1e-8,
    epsrel=1e-8,
):
    """Compute the current the NS or NSN junction with the bias
    V_bias applied to the normal lead form which electrons
    are injected.

    Parameters:
    ----------
    V_bias: float
        The bias applied between the source lead and the superconductor.
    syst: kwant.builder.FiniteSystem
        Finalized Kwant system for the superconducting
        heterostructure.
    params: dict
        System parameters to pass to Kwant solvers. Must include
        the temperature T in units of energy.
    nonloc: bool
        Whether to compute the local (True) or nonlocal (False)
        current.
    in_lead: int
        Normal lead from which electrons are injected, either
        0 or 1.
    out_lead: int
        Normal lead where the conductance is computed, either
        0 or 1.
    biased_lead: int
        Normal lead where bias is applied, either 0 or 1.
    epsabs: float
        Absolute tolerance for the numerical integration.
    epsrel: float
        Relative tolerance for the numerical integration.

    Notes
    ------
    The nonloc option is only compatible with an NSN junction, i.e.
    where the superconductor is sandwiched between two normal leads,
    and will otherwise give wrong results.
    """

    # To calculate the non-local conductance we follow the convention in
    # arXiv:1905.05438, i.e. assume the bias is always applied one lead
    # and the current is measured on the other.
    if nonloc:
        _params = {**params, **{"V_bias_left": 0, "V_bias_right": V_bias}}
    elif len(syst.leads) == 2:
        _params = {**params, **{"V_bias_left": V_bias, "V_bias_right": 0}}
    elif len(syst.leads) == 1:
        _params = {**params, **{"V_bias_left": V_bias}}
    else:
        raise NotImplementedError(
            "Can only compute current in systems\
                                  with 1 or 2 leads."
        )

    T = _params["T"]

    integrand = lambda E: andreev_conductance(
        E, syst, _params, nonloc=nonloc, in_lead=in_lead, out_lead=out_lead
    ) *  (fermi_dist(E, T) - fermi_dist(E + V_bias, T))

    # Temperature-dependent margins for the integration bounds.
    # At |E - mu| = 5*T, the Fermi function is < 0.7% of maximum value.
    lower, upper = np.sort([0, -V_bias])
    current, error = integrate.quad(
        integrand, lower - 6 * T, upper + 6 * T, epsabs=epsabs, epsrel=epsrel
    )

    return current, error


def gap_finder(lead, params, start_Es=(0, 1.0), eps=1e-10, max_iter=200):
    """Compute the induced gap from the dispersion of a lead by finding the
    smallest energy at which there are propagating modes. Positive energies
    are assumed.

    If there are propagating modes at the specified lower energy bound, the
    system is assumed gapless and the lower bound is returned.

    If there are no propagating modes at either the upper or the lower energy
    bounds, we interpret the gap as larger than the upper bound and just return
    the upper bound.

    Parameters:
    -------------------
    lead: kwant.system.InfiniteSystem object
          A finalized Kwant 1D translationally invariant system.
    params: dict
          A container with all Hamiltonian parameters.
    start_Es: tuple of two floats
          Starting guess for an energy interval around the induced gap.
          It is assumed that start_Es[1] > start_Es[0].
    eps: float
          Tolerance for the induced gap. """

    E0, E1 = start_Es
    nmodes0 = lead.modes(energy=E0, params=params)[1].nmodes
    nmodes1 = lead.modes(energy=E1, params=params)[1].nmodes
    if nmodes0 != 0:
        return E0
    if nmodes0 == 0 and nmodes1 == 0:
        # Try to change E1 slightly, in case we we hit
        # an energy range with no propagating modes
        E1 = E1 * 1.05
        nmodes1 = lead.modes(energy=E1, params=params)[1].nmodes
        if nmodes1 == 0:
            return E1
    if not (nmodes0 == 0 and nmodes1 != 0):
        raise ValueError(
            "Give initial energies such that modes are \
            active at one but not the other"
        )
    iterations = 0
    while True:
        iterations += 1
        new_E = 0.5 * (E1 + E0)
        new_nmodes = lead.modes(energy=new_E, params=params)[1].nmodes
        if new_nmodes == 0:
            previous_E = E0
            E0 = new_E
        elif new_nmodes != 0:
            previous_E = E1
            E1 = new_E
        if np.abs(new_E - previous_E) < eps:
            break
        if iterations > max_iter:
            return None
    return new_E


def learner_data(learner):
    if isinstance(learner, adaptive.Learner1D) or isinstance(
        learner, adaptive.Learner2D
    ):
        x, y = learner.data.keys(), learner.data.values()
        x, y = np.array(list(x)), np.array(list(y))
        asrt = np.argsort(x)
        return x[asrt], y[asrt]
    elif isinstance(learner, adaptive.SequenceLearner):
        x, y = learner.data.keys(), learner.data.values()
        x, y = np.array(list(x)), np.array(list(y))
        asrt = np.argsort(x)
        return np.array(learner.sequence)[asrt], y[asrt]
    else:
        raise NotImplementedError
